# pylint: disable=line-too-long
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from unittest.mock import patch

import os
import yaml
from testfixtures import TempDirectory

from ploigos_step_runner.__main__ import main

from tests.helpers.base_test_case import BaseTestCase
from tests.helpers.test_utils import create_sops_side_effect


class TestMain(BaseTestCase):
    def _run_main_test(
        self,
        argv,
        expected_exit_code=None,
        config_files=None,
        expected_results=None
    ):
        cwd = os.getcwd()

        if argv is None:
            argv = []

        try:
            with TempDirectory() as temp_dir:
                os.chdir(temp_dir.path)

                if config_files:
                    argv.append('--config')
                    for config_file in config_files:
                        if isinstance(config_file, dict):
                            config_file_name = config_file['name']
                            config_file_contents = config_file['contents']

                            temp_dir.write(config_file_name, bytes(config_file_contents, 'utf-8'))
                            config_file_path = os.path.join(temp_dir.path, config_file_name)
                            argv.append(config_file_path)
                        else:
                            argv.append(config_file)

                if expected_exit_code is not None:
                    with self.assertRaisesRegex(SystemExit, f"{expected_exit_code}") as cm:
                        main(argv)
                else:
                    main(argv)

                    if expected_results:
                        work_dir_path = os.path.join(temp_dir.path, 'step-runner-working')
                        with open(os.path.join(work_dir_path, "step-runner-results.yml"), 'r') as step_results_file:
                            results = yaml.safe_load(step_results_file.read())
                            self.assertCountEqual(results, expected_results)
        finally:
            os.chdir(cwd)

    def test_init(self):
        """
        Notes
        -----
        See https://medium.com/opsops/how-to-test-if-name-main-1928367290cb
        """
        from ploigos_step_runner import __main__
        with patch.object(__main__, "main", return_value=42):
            with patch.object(__main__, "__name__", "__main__"):
                with patch.object(__main__.sys, 'exit') as mock_exit:
                    __main__.init()
                    assert mock_exit.call_args[0][0] == 42

    def test_no_arguments(self):
        self._run_main_test(None, 2)

    def test_help(self):
        self._run_main_test(['--help'], 0)

    def test_bad_arg(self):
        self._run_main_test(['--bad-arg'], 2)

    def test_config_file_does_not_exist(self):
        self._run_main_test(['--step', 'generate-metadata', '--config', 'does-not-exist.yml'], 101)

    def test_config_file_not_json_or_yaml(self):
        self._run_main_test(['--step', 'generate-metadata'], 102, [
            {
                'name': 'step-runner-config.yaml',
                'contents': ": blarg this: is {} bad syntax"
            }]
                            )

    def test_config_file_no_root_config_key(self):
        self._run_main_test(['--step', 'generate-metadata'], 102, [
            {
                'name': 'step-runner-config.yaml',
                'contents': '{}'
            }]
                            )

    def test_config_file_valid_yaml(self):
        self._run_main_test(['--step', 'foo'], None, [
            {
                'name': 'step-runner-config.yaml',
                'contents': '''---
                step-runner-config:
                    foo:
                        implementer: 'tests.helpers.sample_step_implementers.FooStepImplementer'
                '''
            }]
                            )

    def test_config_file_valid_json(self):
        self._run_main_test(['--step', 'foo'], None, [
            {
                'name': 'step-runner-config.json',
                'contents': '''
                {
                    "step-runner-config": {
                        "foo": {
                        "implementer": 'tests.helpers.sample_step_implementers.FooStepImplementer'
                        }
                    }
                }
                '''
            }]
                            )

    def test_required_step_config_missing(self):
        self._run_main_test(['--step', 'required-step-config-test'], 300, [
            {
                'name': 'step-runner-config.yaml',
                'contents': '''---
                step-runner-config: {}
                '''
            }]
                            )

    def test_required_step_config_pass_via_config_file(self):
        self._run_main_test(['--step', 'required-step-config-test'], None, [
            {
                'name': 'step-runner-config.yaml',
                'contents': '''---
                    step-runner-config:
                        required-step-config-test:
                            implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                            config:
                                required-config-key: "hello world"
                '''
            }]
                            )

    def test_required_step_config_pass_via_runtime_arg_missing(self):
        self._run_main_test(
            [
                '--step', 'required-runtime-step-config-test',
                '--step-config', 'wrong-config="hello world"'
            ],
            200,
            [
                {
                    'name': 'step-runner-config.yaml',
                    'contents': '''---
                    step-runner-config:
                        'required-runtime-step-config-test':
                             implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                             config: {}
                    '''
                }
            ]
        )

    def test_required_step_config_pass_via_runtime_arg_valid(self):
        self._run_main_test(
            [
                '--step', 'required-step-config-test',
                '--step-config', 'required-config-key="hello world"'
            ],
            None,
            [
                {
                    'name': 'step-runner-config.yaml',
                    'contents': '''---
                    step-runner-config:
                        'required-step-config-test':
                             implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                             config: {}
                    '''
                }
            ]
        )

    def test_multiple_config_files_verify_required_key_not_overwritten(self):
        self._run_main_test(
            ['--step', 'required-step-config-test'],
            None,
            [
                {
                    'name': 'step-runner-config1.yaml',
                    'contents': '''---
                        step-runner-config:
                            required-step-config-test:
                                implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                                config:
                                    required-config-key: "hello world"
                                    bar: 'test'
                    '''
                },
                {
                    'name': 'step-runner-config2.yaml',
                    'contents': '''---
                        step-runner-config:
                            required-step-config-test:
                                implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                                config:
                                    key2: "goodbye world"
                    '''
                }
            ]
        )

    def test_multiple_config_files_verify_expected_merge(self):
        self._run_main_test(
            ['--step', 'write-config-as-results'],
            None,
            [
                {
                    'name': 'step-runner-config1.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    key1: "value1"
                                    key2: "value1"
                    '''
                },
                {
                    'name': 'step-runner-config2.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    key3: "value2"
                                    required-config-key: 'value'
                    '''
                }
            ],
            {
                'step-runner-results': {
                    'write-config-as-results': {
                        'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer': {
                            'artifacts': [
                                {'name': 'key1', 'description': '', 'value': 'value1'},
                                {'name': 'key2', 'description': '', 'value': 'value1'},
                                {'name': 'key3', 'description': '', 'value': 'value2'},
                                {'name': 'required-config-key', 'description': '', 'value': 'value'}
                            ],
                            'message': '',
                            'sub-step-implementer-name':
                                'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer',
                            'success': True
                        }
                    }
                }
            }
        )

    def test_multiple_config_files_dup_keys_error(self):
        self._run_main_test(
            ['--step', 'required-step-config-test'],
            102,
            [
                {
                    'name': 'step-runner-config1.yaml',
                    'contents': '''---
                        step-runner-config:
                            required-step-config-test:
                                implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                                config:
                                    required-config-key: "hello world"
                                    bar: 'test'
                    '''
                },
                {
                    'name': 'step-runner-config2.yaml',
                    'contents': '''---
                        step-runner-config:
                            required-step-config-test:
                                implementer: 'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer'
                                config:
                                    bar: "goodbye world"
                    '''
                }
            ]
        )

    def test_multiple_config_files_from_dir(self):
        args = ['--step', 'write-config-as-results']

        with TempDirectory() as temp_dir:
            config_files = [
                {
                    'name': 'step-runner-config1.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    required-config-key: 'value'
                    '''
                },
                {
                    'name': 'foo/a.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    keya: "a"

                    '''
                },
                {
                    'name': 'foo/b.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    keyb: "b"
                    '''
                },
                {
                    'name': 'foo/bar/c.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    keyc: "c"
                    '''
                },
                {
                    'name': 'foo/bar2/c2.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    keyc2: "c2"
                    '''
                },
                {
                    'name': 'foo/bar/meh/d.yaml',
                    'contents': '''---
                        step-runner-config:
                            write-config-as-results:
                                implementer: 'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer'
                                config:
                                    keyd: "d"
                    '''
                }
            ]
            for config_file in config_files:
                config_file_name = config_file['name']
                config_file_contents = config_file['contents']

                temp_dir.write(config_file_name, bytes(config_file_contents, 'utf-8'))

            args.append('--config')
            args.append(os.path.join(temp_dir.path, 'foo'))
            args.append(os.path.join(temp_dir.path, 'step-runner-config1.yaml'))
            self._run_main_test(
                args,
                None,
                None,
                {
                    'step-runner-results': {
                        'write-config-as-results': {
                            'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer': {
                                'artifacts': [
                                    {'name': 'keyc2', 'description': '', 'value': 'c2'},
                                    {'name': 'keya', 'description': '', 'value': 'a'},
                                    {'name': 'keyd', 'description': '', 'value': 'd'},
                                    {'name': 'keyc', 'description': '', 'value': 'c'},
                                    {'name': 'keyb', 'description': '', 'value': 'b'},
                                    {'name': 'required-config-key', 'description': '', 'value': 'value'}
                                ],
                                'message': '',
                                'sub-step-implementer-name':
                                    'tests.helpers.sample_step_implementers.WriteConfigAsResultsStepImplementer',
                                'success': True
                            }
                        }
                    }
                }
            )

    def test_encrypted_value_no_decryptor(self):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config.yml'
        )

        self._run_main_test(
            argv=[
                '--step', 'required-step-config-test',
                '--environment', 'DEV'
            ],
            config_files=[encrypted_config_file_path, config_file_path],
            expected_results={
                'step-runner-results': {
                    'DEV': {
                        'required-step-config-test': {
                            'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer': {
                                'artifacts':[
                                    {'name': 'environment-name', 'description': '', 'value': 'DEV'},
                                    {
                                        'name': 'kube-api-token',
                                        'description': '',
                                        'value': 'ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]'
                                    },
                                    {
                                        'name': 'required-config-key',
                                        'description': '',
                                        'value': 'ENC[AES256_GCM,data:McsZ87srP8gCRNDOysExE/XJ6OaCGyAT3lmNcPXnNvwrucMrBQ==,iv:0cmnMa3tRDaHHdRekzUR57KgGj9fdCLGnWpD+1TUAyM=,tag:svFAjgdBI+mmqopwgKlRFg==,type:str]'
                                    }
                                ],
                                'message': '',
                                'sub-step-implementer-name':
                                    'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer',
                                'success': True
                            }
                        }
                    }
                }
            }
        )

    @patch('sh.sops', create=True)
    def test_encrypted_value_with_sops_decryptor(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config.yml'
        )

        decryptors_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-decryptors.yml'
        )

        mock_decrypted_value = 'mock decrypted value'
        sops_mock.side_effect = create_sops_side_effect(mock_decrypted_value)
        self._run_main_test(
            argv=[
                '--step', 'required-step-config-test',
                '--environment', 'DEV'
            ],
            config_files=[encrypted_config_file_path, config_file_path, decryptors_config_file_path],
            expected_results={
                'step-runner-results': {
                    'DEV': {
                        'required-step-config-test': {
                            'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer': {
                                'artifacts': [
                                    {'name': 'environment-name', 'description': '', 'value': 'DEV'},
                                    {'name': 'kube-api-token', 'description': '', 'value': 'mock decrypted value'},
                                    {'name': 'required-config-key', 'description': '', 'value': 'mock decrypted value'}
                                ],
                                'message': '',
                                'sub-step-implementer-name':
                                    'tests.helpers.sample_step_implementers.RequiredStepConfigStepImplementer',
                                'success': True}
                        }
                    }
                }
            }
        )

    def test_fail(self):
        self._run_main_test(['--step', 'foo'], 200, [
            {
                'name': 'step-runner-config.yaml',
                'contents': '''---
                step-runner-config:
                    foo:
                        implementer: 'tests.helpers.sample_step_implementers.FailStepImplementer'
                '''
            }]
                            )



"""`StepImplementer` for the `validate-environment-configuration` step using config-lint against
a given yaml file to lint.

Executes the config-lint against yml files for user-defined rules. The inputs to this step include:

  - Rules defined by the user in a specified file:
    * Reference:  https://stelligent.github.io/config-lint/#/
  - File path of yml files to lint
    * Specify as a config value, OR
    * Use results from previous step such as ConfilintFromArgocd

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key       | Required | Default               | Description
------------------------|----------|-----------------------|-----------------------------------
`rules`                 | Yes      | ./config_lint.rules   | File containing user-defined rules
`configlint-yml-file`   | Yes      | None                  | File to be linted

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key     | Description
------------------------|------------
`configlint-yml-file`   | File that was linted
`configlint-result-set` | Result of configlint in a text file


Examples
--------
**Example: Step Configuration (minimal)**

    validate-environment-configuration:
    - implementer: ConfiglintFromArgocd
    - implementer: Configlint
      config:
        rules: 'config_lint.rules'

**Example: Generated Config Lint Call (runtime)**

    config-lint -verbose
        -rules rules.yml /home/user/step-runner-working/deploy/file.yml

**Example: Existing Rules File (minimal)**

    version: 1
    description: Rules for Kubernetes spec files
    type: Kubernetes
    files:
      - "*.yml"
    rules:
    - id: EXAMPLE
      severity: FAILURE
      message: Deployment must have istio
      resource: Deployment
      assertions:
        - key: spec.template.metadata.annotations
          op: contains
          value: '"sidecar.istio.io/inject": false'

.. Note:: The configuration examples above uses the ConfiglintFromArgocd step to \
provide the configlint-yml-path.
"""

import os
import sys

import sh
from ploigos_step_runner import StepImplementer
from ploigos_step_runner import StepResult
from ploigos_step_runner.utils.io import create_sh_redirect_to_multiple_streams_fn_callback

DEFAULT_CONFIG = {
    'rules': './config-lint.rules'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'configlint-yml-path'
]


class Configlint(StepImplementer):
    """`StepImplementer` for the validate-environment-configuration step using config-lint against
    a given yaml file to lint.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # configlint-yml-path is required
        configlint_yml_path = self.get_value('configlint-yml-path')

        if not os.path.exists(configlint_yml_path):
            step_result.success = False
            step_result.message = 'File specified in ' \
                                  f'configlint-yml-path not found: {configlint_yml_path}'
            return step_result

        # Required: rules and exists
        rules_file = self.get_value('rules')
        if not os.path.exists(rules_file):
            step_result.success = False
            step_result.message = f'File specified in rules not found: {rules_file}'
            return step_result

        configlint_results_file_path = self.write_working_file('configlint_results_file.txt')
        try:
            # run config-lint writing stdout and stderr to the standard streams
            # as well as to a results file.
            with open(configlint_results_file_path, 'w', encoding='utf-8') \
                    as configlint_results_file:
                out_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                    sys.stdout,
                    configlint_results_file
                ])
                err_callback = create_sh_redirect_to_multiple_streams_fn_callback([
                    sys.stderr,
                    configlint_results_file
                ])

                sh.config_lint(  # pylint: disable=no-member
                    "-verbose",
                    "-debug",
                    "-rules",
                    rules_file,
                    configlint_yml_path,
                    _encoding='UTF-8',
                    _out=out_callback,
                    _err=err_callback,
                    _tee='err'
                )
        except sh.ErrorReturnCode_255:  # pylint: disable=no-member
            # NOTE: expected failure condition,
            #       aka, the config lint run, but found an issue
            #       stderr/stdout is captured in configlint_results_file_path
            step_result.success = False
            step_result.message = 'Failed config-lint scan.'
        except sh.ErrorReturnCode:
            # NOTE: un-expected failure condition
            #       aka, the config lint failed to run for some reason
            #       stderr/stdout is captured in configlint_results_file_path
            step_result.success = False
            step_result.message = 'Unexpected Error invoking config-lint.'

        step_result.add_artifact(
            name='configlint-result-set',
            value=configlint_results_file_path
        )
        step_result.add_artifact(
            name='configlint-yml-path',
            value=configlint_yml_path
        )
        return step_result



"""Test for npm.py

Test for the utility for npm operations.
"""
from unittest.mock import patch, call, mock_open
from io import BytesIO, StringIO, TextIOWrapper

from ploigos_step_runner import StepRunnerException
from tests.helpers.base_test_case import BaseTestCase
from ploigos_step_runner.utils.npm import run_npm
from tests.helpers.test_utils import Any
from pytest import raises
from sh import ErrorReturnCode

class TestNpmUtils(BaseTestCase):

    @patch('sh.npm', create=True) # Given a shell command, 'npm'
    @patch('ploigos_step_runner.utils.io.create_sh_redirect_to_multiple_streams_fn_callback')  # Given a callback that redirects stdio/stderr
    @patch("builtins.open", new_callable=mock_open) # Given that I can open files
    def test_run_script_shell_argument(self, mock_open, redirect_callback_mock, npm_shell_command_mock):

        # When I use run_npm() to run 'command with:args'
        run_npm('/my/output/file', 'mycommand')

        # Then it should run a shell command, `npm command with:args`
        npm_shell_command_mock.assert_any_call(
            'mycommand',
            _out=Any(StringIO),
            _err=Any(StringIO)
        )

    @patch('sh.npm', create=True) # Given a shell command, 'npm'
    @patch('ploigos_step_runner.utils.io.create_sh_redirect_to_multiple_streams_fn_callback')  # Given a callback that redirects stdio/stderr
    @patch("builtins.open", new_callable=mock_open) # Given that I can open files
    def test_run_script_nonzero_exit_code(self, mock_open, redirect_callback_mock, npm_shell_command_mock):

        # Given the shell command exits with an error code
        npm_shell_command_mock.side_effect = ErrorReturnCode('npm', b'mock stdout', b'mock error')

        # When I use run_npm() to run 'mycommand'
        with raises(StepRunnerException): # Then it should raise an exception
            run_npm('/my/output/file', 'mycommand')




import http
import os
import urllib
from unittest.mock import Mock, patch

from ploigos_step_runner.utils.file import (
    base64_encode, create_parent_dir,
    download_source_to_destination,
    download_and_decompress_source_to_destination, get_file_hash,
    parse_yaml_or_json_file, upload_file)
from testfixtures import TempDirectory
from tests.helpers.base_test_case import BaseTestCase


class TestParseYAMLOrJASONFile(BaseTestCase):
    def test_import_yaml(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )
        sample_dict = parse_yaml_or_json_file(sample_file_path)

        self.assertTrue(isinstance(sample_dict, dict))
        self.assertIsNotNone(sample_dict['step-runner-config'])
        self.assertEqual(sample_dict['step-runner-config']['global-defaults']['service-name'], 'fruit')
        self.assertEqual(sample_dict['step-runner-config']['package'][0]['implementer'], 'Maven')

    def test_import_json(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.json'
        )
        sample_dict = parse_yaml_or_json_file(sample_file_path)

        self.assertTrue(isinstance(sample_dict, dict))
        self.assertIsNotNone(sample_dict['step-runner-config'])
        self.assertEqual(sample_dict['step-runner-config']['global-defaults']['service-name'], 'fruit')
        self.assertEqual(sample_dict['step-runner-config']['package'][0]['implementer'], 'Maven')

    def test_import_bad(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'bad.yaml'
        )

        with self.assertRaisesRegex(
            ValueError,
            r"Error parsing file \(.+\) as YAML or JSON:"
        ):
            parse_yaml_or_json_file(sample_file_path)

class TestDownloadAndDecompressSourceToDestination(BaseTestCase):
    def test_https_bz2(self):
        with TempDirectory() as test_dir:

            destination_path = download_and_decompress_source_to_destination(
                source_uri="https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/rhel-8.ds.xml$')
            with open(destination_path) as downloaded_file:
                self.assertTrue(downloaded_file.read())

    def test_https_xml(self):
        with TempDirectory() as test_dir:

            destination_path = download_and_decompress_source_to_destination(
                source_uri="https://www.redhat.com/security/data/cvrf/2020/cvrf-rhba-2020-0017.xml",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/cvrf-rhba-2020-0017.xml$')
            with open(destination_path) as downloaded_file:
                self.assertTrue(downloaded_file.read())

    def test_local_file_download_file_prefix(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'cvrf-rhba-2020-0017.xml'
        )

        with TempDirectory() as test_dir:
            destination_path = download_and_decompress_source_to_destination(
                source_uri=f"file://{sample_file_path}",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/cvrf-rhba-2020-0017.xml$')
            with open(destination_path) as downloaded_file, open(sample_file_path) as sample_file:
                downloaded_file_contents = downloaded_file.read()
                self.assertTrue(downloaded_file_contents)
                self.assertEqual(downloaded_file_contents, sample_file.read())

    def test_local_file_download_forward_slash_prefix(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'cvrf-rhba-2020-0017.xml'
        )

        with TempDirectory() as test_dir:
            destination_path = download_and_decompress_source_to_destination(
                source_uri=f"{sample_file_path}",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/cvrf-rhba-2020-0017.xml$')
            with open(destination_path) as downloaded_file, open(sample_file_path) as sample_file:
                downloaded_file_contents = downloaded_file.read()
                self.assertTrue(downloaded_file_contents)
                self.assertEqual(downloaded_file_contents, sample_file.read())

    def test_bad_protocol(self):
        with TempDirectory() as test_dir:

            with self.assertRaisesRegex(
                ValueError,
                r"Unexpected error, should have been caught by step validation."
                r" Source \(.+\) must start with known protocol \(/|file://\|http://\|https://\)."
            ):
                download_and_decompress_source_to_destination(
                    source_uri="bad://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2",
                    destination_dir=test_dir.path
                )

    def test_https_bad_uri(self):
        with TempDirectory() as test_dir:
            with self.assertRaisesRegex(
                RuntimeError,
                r"Error downloading file \(.+\): HTTP Error 404: Not Found"
            ):
                download_and_decompress_source_to_destination(
                    source_uri="https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/does-not-exist.ds.xml.bz2",
                    destination_dir=test_dir.path
                )

class TestDownloadSourceToDestination(BaseTestCase):
    def test_https_xml(self):
        with TempDirectory() as test_dir:
            destination_path = download_source_to_destination(
                source_uri="https://www.redhat.com/security/data/cvrf/2020/cvrf-rhba-2020-0017.xml",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/cvrf-rhba-2020-0017.xml$')
            with open(destination_path) as downloaded_file:
                self.assertTrue(downloaded_file.read())

    def test_local_file_download_file_prefix(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'cvrf-rhba-2020-0017.xml'
        )

        with TempDirectory() as test_dir:
            destination_path = download_source_to_destination(
                source_uri=f"file://{sample_file_path}",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/cvrf-rhba-2020-0017.xml$')
            with open(destination_path) as downloaded_file, open(sample_file_path) as sample_file:
                downloaded_file_contents = downloaded_file.read()
                self.assertTrue(downloaded_file_contents)
                self.assertEqual(downloaded_file_contents, sample_file.read())

    def test_local_file_download_forward_slash_prefix(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'cvrf-rhba-2020-0017.xml'
        )

        with TempDirectory() as test_dir:
            destination_path = download_source_to_destination(
                source_uri=f"{sample_file_path}",
                destination_dir=test_dir.path
            )

            self.assertIsNotNone(destination_path)
            self.assertRegex(destination_path, rf'{test_dir.path}/cvrf-rhba-2020-0017.xml$')
            with open(destination_path) as downloaded_file, open(sample_file_path) as sample_file:
                downloaded_file_contents = downloaded_file.read()
                self.assertTrue(downloaded_file_contents)
                self.assertEqual(downloaded_file_contents, sample_file.read())

    def test_bad_protocol(self):
        with TempDirectory() as test_dir:

            with self.assertRaisesRegex(
                ValueError,
                r"Unexpected error, should have been caught by step validation."
                r" Source \(.+\) must start with known protocol \(/|file://\|http://\|https://\)."
            ):
                download_source_to_destination(
                    source_uri="bad://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2",
                    destination_dir=test_dir.path
                )

    def test_https_bad_uri(self):
        with TempDirectory() as test_dir:
            with self.assertRaisesRegex(
                RuntimeError,
                r"Error downloading file \(.+\): HTTP Error 404: Not Found"
            ):
                download_source_to_destination(
                    source_uri="https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/does-not-exist.ds.xml.bz2",
                    destination_dir=test_dir.path
                )
class TestUploadFile(BaseTestCase):
    def __create_http_response_side_effect(self, read_return):
        def http_response_side_effect(request):
            mock_response = Mock()
            mock_response.read.return_value = read_return
            mock_response.status = '201'
            mock_response.reason = 'Created'
            return mock_response

        return http_response_side_effect

    def test_source_file_does_not_exist(self):
        with self.assertRaisesRegex(
            ValueError,
            r'Given file path \(/does/not/exist\) to upload does not exist.'
        ):
            upload_file(
                file_path='/does/not/exist',
                destination_uri='/does/not/matter'
            )

    def test_invalid_destination_uri(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        with self.assertRaisesRegex(
            ValueError,
            r'Unexpected error, should have been caught by step validation.'
            r' Destination \(wont-work:///does/not/matter\) must start with known protocol'
            r'\(/|file://|http://|https://\).'
        ):
            upload_file(
                file_path=sample_file_path,
                destination_uri='wont-work:///does/not/matter'
            )

    def test_file_prefix(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        with TempDirectory() as test_dir:
            destination_dir_path = os.path.join(test_dir.path, 'test_dest')
            actual_result = upload_file(
                file_path=sample_file_path,
                destination_uri=f"file://{destination_dir_path}"
            )

            expected_result = os.path.join(destination_dir_path, 'sample.yaml')
            self.assertEqual(expected_result, actual_result)

            with \
                open(actual_result, 'r') as uploaded_file, \
                open(sample_file_path, 'r') as original_file:

                self.assertEqual(original_file.read(), uploaded_file.read())

    def test_forward_slash_prefix(self):
        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        with TempDirectory() as test_dir:
            destination_dir_path = os.path.join(test_dir.path, 'test_dest')
            actual_result = upload_file(
                file_path=sample_file_path,
                destination_uri=f"{destination_dir_path}"
            )

            expected_result = os.path.join(destination_dir_path, 'sample.yaml')
            self.assertEqual(expected_result, actual_result)

            with \
                open(actual_result, 'r') as uploaded_file, \
                open(sample_file_path, 'r') as original_file:

                self.assertEqual(original_file.read(), uploaded_file.read())

    @patch.object(
         urllib.request.OpenerDirector,
         'open'
    )
    def test_http_prefix_no_auth(self, opener_open_mock):
        opener_open_mock.side_effect = self.__create_http_response_side_effect(b'hello world 42')

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        actual_result = upload_file(
            file_path=sample_file_path,
            destination_uri="http://ploigos.com/test/foo42"
        )
        self.assertEqual('status=201, reason=Created, body=hello world 42', actual_result)

    @patch.object(
         urllib.request.OpenerDirector,
         'open'
    )
    def test_https_prefix_no_auth(self, opener_open_mock):
        opener_open_mock.side_effect = self.__create_http_response_side_effect(b'hello world 42')

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        actual_result = upload_file(
            file_path=sample_file_path,
            destination_uri="https://ploigos.com/test/foo42"
        )
        self.assertEqual('status=201, reason=Created, body=hello world 42', actual_result)

    @patch.object(
         urllib.request.OpenerDirector,
         'open'
    )
    @patch('urllib.request.HTTPPasswordMgrWithDefaultRealm')
    def test_https_prefix_auth(self, pass_manager_mock, opener_open_mock):
        opener_open_mock.side_effect = self.__create_http_response_side_effect(b'hello world 42')

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        actual_result = upload_file(
            file_path=sample_file_path,
            destination_uri="https://ploigos.com/test/foo42",
            username='test_user',
            password='pass123'
        )
        self.assertEqual('status=201, reason=Created, body=hello world 42', actual_result)

        pass_manager_mock().add_password.assert_called_once_with(
            None,
            'ploigos.com',
            'test_user',
            'pass123'
        )

    @patch.object(
         urllib.request.OpenerDirector,
         'open'
    )
    def test_https_prefix_http_error(self, opener_open_mock):
        opener_open_mock.side_effect = urllib.error.HTTPError(
            url='https://ploigos.com/test/foo42',
            code=442,
            msg='mock error',
            hdrs=None,
            fp=None
        )

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.yaml'
        )

        with self.assertRaisesRegex(
            RuntimeError,
            rf'Error uploading file \({sample_file_path}\)'
            r' to destination \(https://ploigos.com/test/foo42\) with user name'
            r' \(None\) and password \(None\): HTTP Error 442: mock error'
        ):
            upload_file(
                file_path=sample_file_path,
                destination_uri="https://ploigos.com/test/foo42"
            )
class TestFileMisc(BaseTestCase):
    def test_create_parent_dir(self):
        with TempDirectory() as test_dir:
            file_path = os.path.join(test_dir.path, 'hello/world/does/not/exit/foo.yml')

            self.assertFalse(os.path.exists(file_path))
            self.assertFalse(os.path.exists(os.path.dirname(file_path)))

            create_parent_dir(file_path)
            self.assertFalse(os.path.exists(file_path))
            self.assertTrue(os.path.exists(os.path.dirname(file_path)))

    def test_base64_encode(self):
        with TempDirectory() as test_dir:
            sample_file_path = os.path.join(
                os.path.dirname(__file__),
                'files',
                'sample.txt'
            )

            result = base64_encode(sample_file_path)
            self.assertEqual(result,
                'c2FtcGxlIHRleHQgZmlsZQ=='
            )

    def test_get_file_hash(self):
        with TempDirectory() as test_dir:
            sample_file_path = os.path.join(
                os.path.dirname(__file__),
                'files',
                'sample.txt'
            )

            result = get_file_hash(sample_file_path)
            self.assertEqual(result, '09daa01246aa5ee9c29f64f644627a0ea83247857dfea2665689e26b166eef47')


"""Representation of a sub step configuration.
"""

import copy

from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.utils.dict import deep_merge


class SubStepConfig:
    """Representation of a sub step configuration.

    Parameters
    ----------
    parent_step_config : StepConfig
        Parent StepConfig to this sub step coonfig.
    sub_step_name : str
        Name of this sub step.
    sub_step_implementer_name : str
        Step implementer for this sub step.
    sub_step_config_dict : dict, optional
        Configuration specific to this sub step.
    sub_step_env_config : dict, optional
        Environment specific configuration specific to this sub step.
    sub_step_contine_sub_steps_on_failure : bool
        True to continue executing other sub steps in current step if this sub step fails.
        False to fail all step execution if this sub step fails.

    Attributes
    ----------
    __parent_step_config : StepConfig
    __sub_step_name : str
    __sub_step_implementer_name : str
    __sub_step_config_dict : dict
    __sub_step_env_config : dict
    """

    def __init__( # pylint: disable=too-many-arguments
        self,
        parent_step_config,
        sub_step_name,
        sub_step_implementer_name,
        sub_step_config_dict=None,
        sub_step_env_config=None,
        sub_step_contine_sub_steps_on_failure=False
    ):

        self.__parent_step_config = parent_step_config
        self.__sub_step_name = sub_step_name
        self.__sub_step_implementer_name = sub_step_implementer_name
        self.__sub_step_contine_sub_steps_on_failure = sub_step_contine_sub_steps_on_failure

        if sub_step_config_dict is None:
            sub_step_config_dict = {}
        self.__sub_step_config_dict = sub_step_config_dict

        if sub_step_env_config is None:
            sub_step_env_config = {}
        self.__sub_step_env_config = sub_step_env_config

    @property
    def parent_config(self):
        """
        Returns
        -------
        Config
            Parent Config.
        """
        return self.parent_step_config.parent_config

    @property
    def parent_step_config(self):
        """
        Returns
        -------
        StepConfig
            Parent StepConfig.
        """
        return self.__parent_step_config

    @property
    def step_name(self):
        """Convenience function to get parent step name.

        Returns
        -------
        str
            Parent step name.
        """
        return self.parent_step_config.step_name

    @property
    def step_config_overrides(self):
        """Convenience function to get step config overrides

        Returns
        -------
        str
            Parent step config overrides
        """
        return self.parent_step_config.step_config_overrides

    @property
    def sub_step_name(self):
        """
        Returns
        -------
        str
            Sub step name.
        """
        return self.__sub_step_name

    @property
    def sub_step_implementer_name(self):
        """
        Returns
        -------
        str
            Step implementer name.
        """
        return self.__sub_step_implementer_name

    @property
    def sub_step_config(self):
        """Get a deep copy of the sub step configuration.

        Returns
        -------
        dict
            Deep copy of the sub step configuration.
        """
        return copy.deepcopy(self.__sub_step_config_dict)

    @property
    def global_defaults(self):
        """Convince function for getting the global defaults from the parent config.

        Returns
        -------
        dict
            Deep copy of the global defaults
        """
        return self.parent_config.global_defaults

    @property
    def sub_step_env_config(self):
        """Gets the environment specific configuration for all environments for this sub step.

        Returns
        -------
        dict
            The environment specific configuration for all environments for this sub step.
        """
        return copy.deepcopy(self.__sub_step_env_config)

    @property
    def sub_step_contine_sub_steps_on_failure(self):
        """Gets whether to continue executing other sub steps that belong to the step that
        this sub step belongs to or not if this sub step fails.

        Returns
        -------
        bool
            True to continue executing other sub steps in current step if this sub step fails.
            False to fail all step execution if this sub step fails.
        """
        return self.__sub_step_contine_sub_steps_on_failure

    def get_global_environment_defaults(self, env):
        """Convince function for getting the global environment defaults from the parent config.

        Parameters
        ----------
        env : str
            Environment to get the global environment configuration for.

        Returns
        -------
        dict
            Deep copy of the global defaults for a given environment
        """
        return self.parent_config.get_global_environment_defaults_for_environment(env)

    def get_sub_step_env_config(self, env):
        """Get the sub step environment configuration for an environment.

        Parameters
        ----------
        env : str
            Environment to get the sub step configuration for.

        Returns
        -------
        dict
            Environment specific sub step configuration.
            Empty dict if no environment specific sub step configuration.
        """
        if env in self.sub_step_env_config:
            sub_step_env_config = self.sub_step_env_config[env]
        else:
            sub_step_env_config = {}

        return sub_step_env_config

    def merge_sub_step_config(self, new_sub_step_config):
        """Merge new sub step configuration into the existing sub step configuration.

        Parameters
        ----------
        new_sub_step_config : dict
            New sub step configuration to merge into the existing sub step configuration.

        Raises
        ------
        ValueError
            If new sub step configuration has duplicative leaf keys to
                existing sub step configuration.
        """

        if new_sub_step_config is not None:
            try:
                self.__sub_step_config_dict = deep_merge(
                    self.sub_step_config,
                    copy.deepcopy(new_sub_step_config)
                )
            except ValueError as error:
                raise ValueError(
                    "Error merging new sub step configuration" +
                    " into existing sub step configuration" +
                    f" for sub step ({self.sub_step_name}) of step ({self.step_name}): {error}"
                ) from error

    def merge_sub_step_env_config(self, new_sub_step_env_config):
        """Merge new sub step environment configuration into the existing
        sub step environment configuration.

        Parameters
        ----------
        new_sub_step_env_config : dict
            New sub step environment configuration to merge into
                the existing sub step environment configuration.

        Raises
        ------
        ValueError
            If new sub step environment configuration has duplicative leaf keys to
                existing sub step environment configuration.
        """

        if new_sub_step_env_config is not None:
            try:
                self.__sub_step_env_config = deep_merge(
                    self.__sub_step_env_config,
                    copy.deepcopy(new_sub_step_env_config)
                )
            except ValueError as error:
                raise ValueError(
                    "Error merging new sub step environment configuration" +
                    " into existing sub step environment configuration" +
                    f" for sub step ({self.sub_step_name}) of step ({self.step_name}): {error}"
                ) from error

    def get_config_value(self, key, environment=None, defaults=None):
        """Get the configuration value for a given configuration key from the
        merged set of configuration sources.

        From least precedence to highest precedence.

            1. defaults
            2. Global Configuration Defaults (self.global_config_defaults)
            3. Global Environment Configuration Defaults (self.global_environment_config_defaults)
            4. Step Configuration ( self.step_config)
            5. Step Environment Configuration (self.step_environment_config)
            6. Step Configuration Runtime Overrides (step_config_runtime_overrides)

        Also See
        --------
        get_copy_of_runtime_step_config

        Parameters
        ----------
        key : str
            Key to get the configuration value for.
        environment : str, optional
            Environment to include the configuration for if running in the context of
            a specific environment.
        defaults : dict, optional
            If no value for the given configuration key found in any of the configuration
            sources then use these defaults as last resort.

        Returns
        -------
        str, int, dict, list, or bool or None
            Value of the given configuration key or None if one does not exist
            for this sub step in the given context with the given defaults.
        """
        runtime_step_config = self.__merge_runtime_step_config(environment, defaults)

        if key in runtime_step_config:
            if isinstance(runtime_step_config[key], ConfigValue):
                value = runtime_step_config[key].value
            else:
                value = ConfigValue.convert_leaves_to_values(runtime_step_config[key])
        else:
            value = None

        return value

    def get_copy_of_runtime_step_config(self, environment=None, defaults=None):
        """Take all of the context about this sub step merges together a single dictionary
        with all of the configuration for a given step.

        From least precedence to highest precedence.

            1. defaults
            2. Global Configuration Defaults (self.global_config_defaults)
            3. Global Environment Configuration Defaults (self.global_environment_config_defaults)
            4. Step Configuration ( self.step_config)
            5. Step Environment Configuration (self.step_environment_config)
            6. Step Configuration Runtime Overrides (step_config_runtime_overrides)

        Also See
        --------
        get_config_value

        Parameters
        ----------
        environment : str, optional
            Environment to get the runtime step configuration for
        defaults : dict, optional
            Defaults to use if no other configuration specified from any other source
            for each given key.

        Returns
        -------
        dict
            A deep copy of the merged runtime step configuration
        """
        defaults = defaults if defaults else {}

        return copy.deepcopy(self.__merge_runtime_step_config(environment, defaults))

    def __merge_runtime_step_config(self, environment=None, defaults=None):
        """Take all of the context about this sub step merges together a single dictionary
        with all of the configuration for a given step.

        From least precedence to highest precedence.

            1. defaults
            2. Global Configuration Defaults (self.global_config_defaults)
            3. Global Environment Configuration Defaults (self.global_environment_config_defaults)
            4. Step Configuration ( self.step_config)
            5. Step Environment Configuration (self.step_environment_config)
            6. Step Configuration Runtime Overrides (step_config_runtime_overrides)

        Notes
        -----
        This is not intended to be accessed outside of this class since it gives direct access
        to the underlying dictionaries which could in theory be changed, which would not be the
        intended use.

        Parameters
        ----------
        environment : str, optional
            Environment to get the runtime step configuration for
        defaults : dict, optional
            Defaults to use if no other configuration specified from any other source
            for each given key.

        Returns
        -------
        dict
            Merged runtime step configuration
        """
        defaults = defaults if defaults else {}

        return {
            **defaults,
            **self.global_defaults,
            **self.get_global_environment_defaults(environment),
            **self.sub_step_config,
            **self.get_sub_step_env_config(environment),
            **self.step_config_overrides,
        }



import os
from pathlib import Path
from shutil import copyfile
from unittest.mock import PropertyMock, patch

from ploigos_step_runner import StepResult, StepRunnerException, WorkflowResult
from ploigos_step_runner.config.config import Config
from ploigos_step_runner.step_implementers.shared.maven_generic import \
    MavenGeneric
from ploigos_step_runner.utils.file import create_parent_dir
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.StepImplementer.__init__")
class TestStepImplementerSharedMavenGeneric___init__(BaseStepImplementerTestCase):

    def test_no_environment_no_maven_phases_and_goals(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        step_implementer = MavenGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        self.assertIsNone(step_implementer._MavenGeneric__maven_settings_file)
        self.assertIsNone(step_implementer._MavenGeneric__maven_phases_and_goals)
        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None
        )

    def test_with_environment_no_maven_phases_and_goals(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        step_implementer = MavenGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='test-env'
        )

        self.assertIsNone(step_implementer._MavenGeneric__maven_settings_file)
        self.assertIsNone(step_implementer._MavenGeneric__maven_phases_and_goals)
        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='test-env'
        )

    def test_no_environment_with_maven_phases_and_goals(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        step_implementer = MavenGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            maven_phases_and_goals=['fake-phase']
        )

        self.assertIsNone(step_implementer._MavenGeneric__maven_settings_file)
        self.assertEqual(
            step_implementer._MavenGeneric__maven_phases_and_goals,
            ['fake-phase']
        )
        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None
        )

class TestStepImplementerSharedMavenGeneric_step_implementer_config_defaults(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenGeneric.step_implementer_config_defaults(),
            {
                'pom-file': 'pom.xml',
                'tls-verify': True,
                'maven-profiles': [],
                'maven-additional-arguments': [],
                'maven-no-transfer-progress': True
            }
        )

class TestStepImplementerSharedMavenGeneric__required_config_or_result_keys(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenGeneric._required_config_or_result_keys(),
            [
                'pom-file',
                'maven-phases-and-goals'
            ]
        )

class BaseTestStepImplementerSharedMavenGeneric(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=MavenGeneric,
            step_config=step_config,
            step_name='foo',
            implementer='MavenGeneric',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

@patch("ploigos_step_runner.StepImplementer._validate_required_config_or_previous_step_result_artifact_keys")
class TestStepImplementerSharedMavenGeneric__validate_required_config_or_previous_step_result_artifact_keys(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_valid(self, mock_super_validate):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-phases-and-goals': 'fake-phase'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            Path(pom_file_path).touch()
            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            mock_super_validate.assert_called_once_with()

    def test_pom_file_does_not_exist(self, mock_super_validate):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = '/does/not/exist/pom.xml'
            step_config = {
                'pom-file': pom_file_path,
                'maven-phases-and-goals': 'fake-phase'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                AssertionError,
                rf'Given maven pom file \(pom-file\) does not exist: {pom_file_path}'
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()
                mock_super_validate.assert_called_once_with()

class TestStepImplementerSharedMavenGeneric_maven_phases_and_goals(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_use_object_property_no_config_value(self):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = None

        step_implementer = MavenGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            maven_phases_and_goals=['fake-phase']
        )

        self.assertEqual(
            step_implementer.maven_phases_and_goals,
            ['fake-phase']
        )

    def test_use_object_property_with_config_value(self):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        step_config = {
            'maven-phases-and-goals': ['config-value-fake-phase']
        }
        config = Config({
            Config.CONFIG_KEY: {
                'foo': [
                    {
                        'implementer': 'MavenGeneric',
                        'config': step_config
                    }
                ]

            }
        })

        step_implementer = MavenGeneric(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            maven_phases_and_goals=['object-property-fake-phase']
        )

        self.assertEqual(
            step_implementer.maven_phases_and_goals,
            ['object-property-fake-phase']
        )

    def test_use_config_value(self):
        parent_work_dir_path = '/fake/path'
        step_config = {
            'maven-phases-and-goals': ['config-value-fake-phase']
        }

        step_implementer = self.create_step_implementer(
            step_config=step_config,
            parent_work_dir_path=parent_work_dir_path,
        )

        self.assertEqual(
            step_implementer.maven_phases_and_goals,
            ['config-value-fake-phase']
        )

@patch(
    "ploigos_step_runner.step_implementers.shared.maven_generic.generate_maven_settings",
    return_value='/mock/settings.xml'
)
class TestStepImplementerSharedMavenGeneric_maven_settings_file(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_no_config(self, mock_gen_mvn_settings):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-phases-and-goals': 'fake-phase'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # call first time
            maven_settings_file = step_implementer.maven_settings_file
            self.assertEqual(
                maven_settings_file,
                '/mock/settings.xml'
            )
            mock_gen_mvn_settings.assert_called_once_with(
                working_dir=step_implementer.work_dir_path,
                maven_servers=None,
                maven_repositories=None,
                maven_mirrors=None
            )

            # call second time
            mock_gen_mvn_settings.reset_mock()
            maven_settings_file = step_implementer.maven_settings_file
            self.assertEqual(
                maven_settings_file,
                '/mock/settings.xml'
            )
            mock_gen_mvn_settings.assert_not_called()

    def test_given_config_maven_servers_maven_repos_and_maven_mirrors(self, mock_gen_mvn_settings):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            maven_servers = {
                'internal-mirror-1': {
                    'id': 'internal-server',
                    'password': 'you-wish',
                    'username': 'team-a'
                }
            }
            maven_repositories = {
                'internal-mirror-1': {
                    'id': 'internal-mirror-1',
                    'url': 'https://nexus.apps.ploigos.com/repository/maven-public/',
                    'snapshots': 'true',
                    'releases': 'true'
                }
            }
            maven_mirrors = {
                'internal-mirror-1': {
                    'id': 'internal-mirror',
                    'mirror-of': '*',
                    'url': 'https://nexus.apps.ploigos.com/repository/maven-public/'
                }
            }

            step_config = {
                'pom-file': pom_file_path,
                'maven-phases-and-goals': 'fake-phase',
                'maven-servers': maven_servers,
                'maven-repositories': maven_repositories,
                'maven-mirrors': maven_mirrors
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # call first time
            maven_settings_file = step_implementer.maven_settings_file
            self.assertEqual(
                maven_settings_file,
                '/mock/settings.xml'
            )
            mock_gen_mvn_settings.assert_called_once_with(
                working_dir=step_implementer.work_dir_path,
                maven_servers=maven_servers,
                maven_repositories=maven_repositories,
                maven_mirrors=maven_mirrors
            )

            # call second time
            mock_gen_mvn_settings.reset_mock()
            maven_settings_file = step_implementer.maven_settings_file
            self.assertEqual(
                maven_settings_file,
                '/mock/settings.xml'
            )
            mock_gen_mvn_settings.assert_not_called()

@patch('ploigos_step_runner.step_implementers.shared.maven_generic.write_effective_pom')
class TestStepImplementerSharedMavenGeneric__get_effective_pom(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_call_once(self, write_effective_pom_mock):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # mock effective pom
            Path(pom_file_path).touch()
            def write_effective_pom_mock_side_effect(pom_file_path, output_path, profiles):
                create_parent_dir(pom_file_path)
                copyfile(pom_file_path, output_path)
            write_effective_pom_mock.side_effect = write_effective_pom_mock_side_effect

            # first call
            expected_effective_pom_path = os.path.join(
                step_implementer.work_dir_path,
                'effective-pom.xml'
            )
            actual_effective_pom_path = step_implementer._get_effective_pom()
            self.assertEqual(actual_effective_pom_path, expected_effective_pom_path)
            write_effective_pom_mock.assert_called_once_with(
                pom_file_path=pom_file_path,
                output_path=expected_effective_pom_path,
                profiles=[]
            )

    def test__get_effective_pom_call_twice(self, write_effective_pom_mock):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # mock effective pom
            Path(pom_file_path).touch()
            def write_effective_pom_mock_side_effect(pom_file_path, output_path, profiles):
                create_parent_dir(pom_file_path)
                copyfile(pom_file_path, output_path)
            write_effective_pom_mock.side_effect = write_effective_pom_mock_side_effect

            # first call
            expected_effective_pom_path = os.path.join(
                step_implementer.work_dir_path,
                'effective-pom.xml'
            )
            actual_effective_pom_path = step_implementer._get_effective_pom()
            self.assertEqual(actual_effective_pom_path, expected_effective_pom_path)
            write_effective_pom_mock.assert_called_once_with(
                pom_file_path=pom_file_path,
                output_path=expected_effective_pom_path,
                profiles=[]
            )

            # second call
            write_effective_pom_mock.reset_mock()
            expected_effective_pom_path = os.path.join(
                step_implementer.work_dir_path,
                'effective-pom.xml'
            )
            actual_effective_pom_path = step_implementer._get_effective_pom()
            self.assertEqual(actual_effective_pom_path, expected_effective_pom_path)
            write_effective_pom_mock.assert_not_called()

    def test_call_with_profiles(self, write_effective_pom_mock):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-profiles': ['mock-profile1']
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # mock effective pom
            Path(pom_file_path).touch()
            def write_effective_pom_mock_side_effect(pom_file_path, output_path, profiles):
                create_parent_dir(pom_file_path)
                copyfile(pom_file_path, output_path)
            write_effective_pom_mock.side_effect = write_effective_pom_mock_side_effect

            # first call
            expected_effective_pom_path = os.path.join(
                step_implementer.work_dir_path,
                'effective-pom.xml'
            )
            actual_effective_pom_path = step_implementer._get_effective_pom()
            self.assertEqual(actual_effective_pom_path, expected_effective_pom_path)
            write_effective_pom_mock.assert_called_once_with(
                pom_file_path=pom_file_path,
                output_path=expected_effective_pom_path,
                profiles=['mock-profile1']
            )

@patch('ploigos_step_runner.step_implementers.shared.maven_generic.get_xml_element_by_path')
@patch.object(MavenGeneric, '_get_effective_pom')
class TestStepImplementerSharedMavenGeneric__get_effective_pom_element(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_result(self, get_effective_pom_mock, get_xml_element_by_path_mock):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            def get_effective_pom_side_effect():
                return '/does/not/matter/pom.xml'
            get_effective_pom_mock.side_effect = get_effective_pom_side_effect

            step_implementer._get_effective_pom_element('foo')
            get_effective_pom_mock.assert_called_once_with()
            get_xml_element_by_path_mock.assert_called_once_with(
                '/does/not/matter/pom.xml',
                'foo',
                default_namespace='mvn'
            )

@patch('ploigos_step_runner.step_implementers.shared.maven_generic.run_maven')
@patch.object(
    MavenGeneric,
    'maven_phases_and_goals',
    new_callable=PropertyMock,
    return_value=['fake-phase']
)
@patch.object(
    MavenGeneric,
    'maven_settings_file',
    new_callable=PropertyMock,
    return_value='/fake/settings.xml'
)
class TestStepImplementerSharedMavenGeneric__run_maven_step(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_defaults(self, mock_settings_file, mock_phases_and_goals, mock_run_maven):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=[],
                pom_file=pom_file_path,
                tls_verify=True,
                profiles=[],
                no_transfer_progress=True,
                settings_file='/fake/settings.xml'
            )


    def test_custom_profile(self, mock_settings_file, mock_phases_and_goals, mock_run_maven):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-profiles': ['fake-profile-1', 'fakse-profile-2']
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=[],
                pom_file=pom_file_path,
                tls_verify=True,
                profiles=['fake-profile-1', 'fakse-profile-2'],
                no_transfer_progress=True,
                settings_file='/fake/settings.xml'
            )

    def test_no_tls_verify(self, mock_settings_file, mock_phases_and_goals, mock_run_maven):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'tls-verify': False
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=[],
                pom_file=pom_file_path,
                tls_verify=False,
                profiles=[],
                no_transfer_progress=True,
                settings_file='/fake/settings.xml'
            )

    def test_yes_transfer_progress(self, mock_settings_file, mock_phases_and_goals, mock_run_maven):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-no-transfer-progress': False
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=[],
                pom_file=pom_file_path,
                tls_verify=True,
                profiles=[],
                no_transfer_progress=False,
                settings_file='/fake/settings.xml'
            )

    def test_config_additional_arguments(
        self,
        mock_settings_file,
        mock_phases_and_goals,
        mock_run_maven
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-additional-arguments': ['-Dfake.config.arg=True']
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=['-Dfake.config.arg=True'],
                pom_file=pom_file_path,
                tls_verify=True,
                profiles=[],
                no_transfer_progress=True,
                settings_file='/fake/settings.xml'
            )

    def test_step_implementer_additional_arguments(
        self,
        mock_settings_file,
        mock_phases_and_goals,
        mock_run_maven
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path,
                step_implementer_additional_arguments=['-Dfake.step_implementer.arg=True']
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=['-Dfake.step_implementer.arg=True'],
                pom_file=pom_file_path,
                tls_verify=True,
                profiles=[],
                no_transfer_progress=True,
                settings_file='/fake/settings.xml'
            )

    def test_step_implementer_additional_arguments_and_config_additional_arguments(
        self,
        mock_settings_file,
        mock_phases_and_goals,
        mock_run_maven
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file_path = os.path.join(test_dir.path, 'pom.xml')
            step_config = {
                'pom-file': pom_file_path,
                'maven-additional-arguments': ['-Dfake.config.arg=True']
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            mvn_output_file_path = os.path.join(test_dir.path, 'maven-output.txt')
            step_implementer._run_maven_step(
                mvn_output_file_path=mvn_output_file_path,
                step_implementer_additional_arguments=['-Dfake.step_implementer.arg=True']
            )

            mock_run_maven.assert_called_with(
                mvn_output_file_path=mvn_output_file_path,
                phases_and_goals=['fake-phase'],
                additional_arguments=['-Dfake.step_implementer.arg=True', '-Dfake.config.arg=True'],
                pom_file=pom_file_path,
                tls_verify=True,
                profiles=[],
                no_transfer_progress=True,
                settings_file='/fake/settings.xml'
            )

@patch.object(MavenGeneric, '_run_maven_step')
@patch.object(MavenGeneric, 'write_working_file', return_value='/mock/mvn_output.txt')
class TestStepImplementerSharedMavenGeneric__run_step(
    BaseTestStepImplementerSharedMavenGeneric
):
    def test_success(self, mock_write_working_file, mock_run_maven_step):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='foo',
                sub_step_name='MavenGeneric',
                sub_step_implementer_name='MavenGeneric'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )

    def test_fail(self, mock_write_working_file, mock_run_maven_step):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            step_config = {}
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step with mock failure
            mock_run_maven_step.side_effect = StepRunnerException('Mock error running maven')
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='foo',
                sub_step_name='MavenGeneric',
                sub_step_implementer_name='MavenGeneric'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value='/mock/mvn_output.txt'
            )
            expected_step_result.message = "Error running maven. " \
                "More details maybe found in 'maven-output' report artifact: "\
                "Mock error running maven"
            expected_step_result.success = False

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called_once()
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_output.txt'
            )



"""Test for maven.py

Test for the utility for maven operations.
"""
import re
import xml.etree.ElementTree as ET
from io import BytesIO, StringIO
from pathlib import Path
from unittest.mock import call, mock_open, patch

import sh
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.maven import *
from testfixtures import TempDirectory
from tests.helpers.base_test_case import BaseTestCase
from tests.helpers.test_utils import Any


class TestMavenUtils_other(BaseTestCase):
    # def test_generate_maven_settings_maven_servers_is_none(self):

    def test_generate_maven_servers_exist(self):
        maven_servers = [
            {
                "id": "one",
                "username": "user1",
                "password": "password1"
            }
        ]

        settings = '''<settings><servers><server><id>one</id><username>user1</username><password>password1</password></server></servers></settings>'''

        with TempDirectory() as temp_dir:
            generate_maven_settings(temp_dir.path, maven_servers, None, None)
            with open(temp_dir.path + '/settings.xml', 'r') as tester:
                results = tester.read()
                self.assertEqual(results, settings)

    def test_generate_maven_servers_id_does_not_exist(self):
        maven_servers = [
            {
                "username": "user1",
                "password": "password1"
            }
        ]
        with TempDirectory() as temp_dir:
            with self.assertRaisesRegex(
                AssertionError,
                r"Configuration for maven servers must specify a 'id':"
                r" {'username': 'user1', 'password': 'password1'}"
            ):
                generate_maven_settings(temp_dir.path, maven_servers, None, None)

    def test_generate_maven_servers_username_does_not_exist(self):
        maven_servers = [
            {
                "id": "one",
                "password": "password1"
            }
        ]
        with TempDirectory() as temp_dir:
            with self.assertRaisesRegex(
                AssertionError,
                r"Configuration for maven servers must specify both "
                r"'username' and 'password' or neither: {'id': 'one', 'password': 'password1'}"
            ):
                generate_maven_settings(temp_dir.path, maven_servers, None, None)

    def test_generate_maven_repositories_exists(self):
        maven_repositories = [
            {
                "id": "repo1",
                "url": "repo_url1",
                "releases": "true",
                "snapshots": "true"
            }
        ]
        settings = '''<settings><profiles><profile><repositories><repository><id>repo1</id><url>repo_url1</url><releases><enabled>true</enabled></releases><snapshots><enabled>true</enabled></snapshots></repository></repositories></profile></profiles></settings>'''

        with TempDirectory() as temp_dir:
            generate_maven_settings(temp_dir.path, None, maven_repositories, None)
            with open(temp_dir.path + '/settings.xml', 'r') as tester:
                results = tester.read()
                self.assertEqual(results, settings)

    def test_generate_maven_repositories_id_does_not_exists(self):
        maven_repositories = [
            {
                "url": "repo_url1",
                "releases": "true",
                "snapshots": "true"
            }
        ]
        with TempDirectory() as temp_dir:
            with self.assertRaisesRegex(
                AssertionError,
                r"Configuration for maven repository must specify a 'id':"
                r" {'url': 'repo_url1', 'releases': 'true', 'snapshots': 'true'}"
            ):
                generate_maven_settings(temp_dir.path, None, maven_repositories, None)

    def test_generate_maven_mirrors_exists(self):
        maven_mirrors = [
            {
                "id": "mirror1",
                "url": "mirror_url1",
                "mirror-of": "false"
            }
        ]

        settings = '''<settings><mirrors><mirror><id>mirror1</id><url>mirror_url1</url><mirrorOf>false</mirrorOf></mirror></mirrors></settings>'''

        with TempDirectory() as temp_dir:
            generate_maven_settings(temp_dir.path, None, None, maven_mirrors)
            with open(temp_dir.path + '/settings.xml', 'r') as tester:
                results = tester.read()
                self.assertEqual(results, settings)

    def test_generate_maven_mirrors_exists_id_does_not_exists(self):
        maven_mirrors = [
            {
                "url": "mirror_url1",
                "mirror-of": "false"
            }
        ]
        with TempDirectory() as temp_dir:
            with self.assertRaisesRegex(
                AssertionError,
                r"Configuration for maven mirror must specify a 'id':"
                r" {'url': 'mirror_url1', 'mirror-of': 'false'}"
            ):
                generate_maven_settings(temp_dir.path, None, None, maven_mirrors)

    def test_generate_maven_params_empty(self):
        settings = '<settings />'

        with TempDirectory() as temp_dir:
            generate_maven_settings(temp_dir.path, None, None, None)
            with open(temp_dir.path + '/settings.xml', 'r') as tester:
                results = tester.read()
                self.assertEqual(results, settings)

    def test_generate_maven_mirrors_empty(self):
        settings = '<settings />'

        with TempDirectory() as temp_dir:
            generate_maven_settings(temp_dir.path, None, None, None)
            with open(temp_dir.path + '/settings.xml', 'r') as tester:
                results = tester.read()
                self.assertEqual(results, settings)

    def test_add_maven_server_no_user_no_pass(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)
        servers_element = ET.Element('servers')
        root_element.append(servers_element)

        add_maven_server(
            parent_element=servers_element,
            maven_server_id='foo'
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers><server><id>foo</id></server></servers></settings>'
        )

    def test_add_maven_server_user_and_pass(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)
        servers_element = ET.Element('servers')
        root_element.append(servers_element)

        add_maven_server(
            parent_element=servers_element,
            maven_server_id='foo',
            maven_server_username='test-user',
            maven_server_password='test-pass'
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers><server><id>foo</id>'
            '<username>test-user</username>'
            '<password>test-pass</password>'
            '</server></servers></settings>'
        )

    def test_add_maven_server_yes_user_no_pass(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)
        servers_element = ET.Element('servers')
        root_element.append(servers_element)

        add_maven_server(
            parent_element=servers_element,
            maven_server_id='foo',
            maven_server_username='test-user'
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers><server><id>foo</id></server></servers></settings>'
        )

    def test_add_maven_server_no_user_yes_pass(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)
        servers_element = ET.Element('servers')
        root_element.append(servers_element)

        add_maven_server(
            parent_element=servers_element,
            maven_server_id='foo',
            maven_server_password='test-pass'
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers><server><id>foo</id></server></servers></settings>'
        )

    def test_add_maven_servers_list(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)

        maven_servers = [
            {
                'id': 'id1'
            },
            {
                'id': 'id1',
                'username': 'username1',
                'password': 'password1'
            }
        ]

        add_maven_servers(
            parent_element=root_element,
            maven_servers=maven_servers
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers>'
            '<server><id>id1</id></server>'
            '<server><id>id1</id><username>username1</username><password>password1</password></server>'
            '</servers></settings>'
        )

    def test_add_maven_servers_dict_no_ids(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)

        maven_servers = {
            'id-from-key-1': {
                'username': 'username1',
                'password': 'password1'
            },
            'id-from-key-2': {
                'username': 'username2',
                'password': 'password2'
            }
        }

        add_maven_servers(
            parent_element=root_element,
            maven_servers=maven_servers
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers>'
            '<server><id>id-from-key-1</id><username>username1</username><password>password1</password></server>'
            '<server><id>id-from-key-2</id><username>username2</username><password>password2</password></server>'
            '</servers></settings>'
        )

    def test_add_maven_servers_dict_with_ids(self):
        root_element = ET.Element('settings')
        tree = ET.ElementTree(root_element)

        maven_servers = {
            'id-from-key-1': {
                'username': 'username1',
                'password': 'password1'
            },
            'id-from-key-2': {
                'id': 'override-id-2',
                'username': 'username2',
                'password': 'password2'
            }
        }

        add_maven_servers(
            parent_element=root_element,
            maven_servers=maven_servers
        )

        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            '<settings><servers>'
            '<server><id>id-from-key-1</id><username>username1</username><password>password1</password></server>'
            '<server><id>override-id-2</id><username>username2</username><password>password2</password></server>'
            '</servers></settings>'
        )

    def test_add_maven_servers_list_missing_pass(self):
        root_element = ET.Element('settings')

        maven_servers = [
            {
                'id': 'invalid',
                'username': 'username1',
            },
            {
                'id': 'valid',
                'username': 'username1',
                'password': 'password1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven servers must specify both 'username' and 'password'"
            r" or neither: {'id': 'invalid', 'username': 'username1'}"
        ):
            add_maven_servers(
                parent_element=root_element,
                maven_servers=maven_servers
            )

    def test_add_maven_servers_list_missing_username(self):
        root_element = ET.Element('settings')

        maven_servers = [
            {
                'id': 'invalid',
                'password': 'password1'
            },
            {
                'id': 'valid',
                'username': 'username1',
                'password': 'password1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven servers must specify both 'username' and 'password'"
            r" or neither: {'id': 'invalid', 'password': 'password1'}"
        ):
            add_maven_servers(
                parent_element=root_element,
                maven_servers=maven_servers
            )
    def test_add_maven_servers_dict_missing_pass(self):
        root_element = ET.Element('settings')

        maven_servers = {
            'invalid': {
                'username': 'username1',
            },
            'valid': {
                'username': 'username1',
                'password': 'password1'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven server \(invalid\) must specify"
            r" both 'username' and 'password' or neither: {'username': 'username1'}"
        ):
            add_maven_servers(
                parent_element=root_element,
                maven_servers=maven_servers
            )

    def test_add_maven_servers_dict_missing_username(self):
        root_element = ET.Element('settings')

        maven_servers = {
            'invalid': {
                'password': 'password1'
            },
            'valid': {
                'username': 'username1',
                'password': 'password1'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven server \(invalid\) must specify"
            r" both 'username' and 'password' or neither: {'password': 'password1'}"
        ):
            add_maven_servers(
                parent_element=root_element,
                maven_servers=maven_servers
            )

    def test_add_maven_repository_no_release_no_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url'
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_repository_false_release_no_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url',
            releases_enabled=False
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url><releases><enabled>False</enabled></releases></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_repository_true_release_no_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url',
            releases_enabled=True
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url><releases><enabled>True</enabled></releases></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_repository_no_release_false_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url',
            snapshots_enabled=False
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url><snapshots><enabled>False</enabled></snapshots></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_repository_no_release_true_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url',
            snapshots_enabled=True
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url><snapshots><enabled>True</enabled></snapshots></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_repository_False_release_true_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url',
            releases_enabled=False,
            snapshots_enabled=True
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url>"
            r"<releases><enabled>False</enabled></releases>"
            r"<snapshots><enabled>True</enabled></snapshots></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_repository_true_release_true_snapshot(self):
        root_element = ET.Element('settings')
        profiles_element = ET.Element('profiles')
        root_element.append(profiles_element)
        profile_element = ET.Element('profile')
        profiles_element.append(profile_element)
        repositories_element = ET.Element('repositories')
        profile_element.append(repositories_element)

        add_maven_repository(
            parent_element=repositories_element,
            repository_id='test-id-1',
            repository_url='test-url',
            releases_enabled=True,
            snapshots_enabled=True
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles>"
            r"<profile><repositories>"
            r"<repository><id>test-id-1</id><url>test-url</url>"
            r"<releases><enabled>True</enabled></releases>"
            r"<snapshots><enabled>True</enabled></snapshots></repository>"
            r"</repositories></profile>"
            r"</profiles></settings>"
        )

    def test_add_maven_servers_list(self):
        root_element = ET.Element('settings')

        maven_repositories = [
            {
                'id': 'server-id-1',
                'url': 'url-1'
            },
            {
                'id': 'server-id-2',
                'url': 'url-2'
            }
        ]

        add_maven_repositories(
            parent_element=root_element,
            maven_repositories=maven_repositories
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles><profile><repositories>"
            r"<repository><id>server-id-1</id><url>url-1</url></repository>"
            r"<repository><id>server-id-2</id><url>url-2</url></repository>"
            r"</repositories></profile></profiles></settings>"
        )

    def test_add_maven_servers_dict_with_id_elements(self):
        root_element = ET.Element('settings')

        maven_repositories = {
            'ignore-me-1': {
                'id': 'server-id-1',
                'url': 'url-1'
            },
            'ingnore-me-2': {
                'id': 'server-id-2',
                'url': 'url-2'
            }
        }

        add_maven_repositories(
            parent_element=root_element,
            maven_repositories=maven_repositories
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles><profile><repositories>"
            r"<repository><id>server-id-1</id><url>url-1</url></repository>"
            r"<repository><id>server-id-2</id><url>url-2</url></repository>"
            r"</repositories></profile></profiles></settings>"
        )

    def test_add_maven_servers_dict_no_id_elements(self):
        root_element = ET.Element('settings')

        maven_repositories = {
            'use-me-1': {
                'url': 'url-1'
            },
            'ingnore-me-2': {
                'id': 'server-id-2',
                'url': 'url-2'
            }
        }

        add_maven_repositories(
            parent_element=root_element,
            maven_repositories=maven_repositories
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><profiles><profile><repositories>"
            r"<repository><id>use-me-1</id><url>url-1</url></repository>"
            r"<repository><id>server-id-2</id><url>url-2</url></repository>"
            r"</repositories></profile></profiles></settings>"
        )

    def test_add_maven_mirror_valid(self):
        root_element = ET.Element('settings')
        mirrors_element = ET.Element('mirrors')
        root_element.append(mirrors_element)

        add_maven_mirror(
            parent_element=mirrors_element,
            maven_mirror_id='test-id-1',
            maven_mirror_url='test-url-1',
            maven_mirror_mirror_of='test-mirror-of-1'
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><mirrors>"
            r"<mirror>"
            r"<id>test-id-1</id>"
            r"<url>test-url-1</url>"
            r"<mirrorOf>test-mirror-of-1</mirrorOf>"
            r"</mirror>"
            r"</mirrors></settings>"
        )

    def test_add_maven_mirrors_list_valid(self):
        root_element = ET.Element('settings')

        maven_mirrors = [
            {
                'id': 'test-id-1',
                'url': 'test-url-1',
                'mirror-of': 'test-mirror-of-1'
            },
            {
                'id': 'test-id-2',
                'url': 'test-url-2',
                'mirror-of': 'test-mirror-of-2'
            }
        ]

        add_maven_mirrors(
            parent_element=root_element,
            maven_mirrors=maven_mirrors
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><mirrors>"
            r"<mirror>"
            r"<id>test-id-1</id>"
            r"<url>test-url-1</url>"
            r"<mirrorOf>test-mirror-of-1</mirrorOf>"
            r"</mirror>"
            r"<mirror>"
            r"<id>test-id-2</id>"
            r"<url>test-url-2</url>"
            r"<mirrorOf>test-mirror-of-2</mirrorOf>"
            r"</mirror>"
            r"</mirrors></settings>"
        )

    def test_add_maven_mirrors_dict_with_ids(self):
        root_element = ET.Element('settings')

        maven_mirrors = {
            'ignore-me-1': {
                'id': 'test-id-1',
                'url': 'test-url-1',
                'mirror-of': 'test-mirror-of-1'
            },
            'ignore-me-2': {
                'id': 'test-id-2',
                'url': 'test-url-2',
                'mirror-of': 'test-mirror-of-2'
            }
        }

        add_maven_mirrors(
            parent_element=root_element,
            maven_mirrors=maven_mirrors
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><mirrors>"
            r"<mirror>"
            r"<id>test-id-1</id>"
            r"<url>test-url-1</url>"
            r"<mirrorOf>test-mirror-of-1</mirrorOf>"
            r"</mirror>"
            r"<mirror>"
            r"<id>test-id-2</id>"
            r"<url>test-url-2</url>"
            r"<mirrorOf>test-mirror-of-2</mirrorOf>"
            r"</mirror>"
            r"</mirrors></settings>"
        )

    def test_add_maven_mirrors_dict_without_id(self):
        root_element = ET.Element('settings')

        maven_mirrors = {
            'use-me-1': {
                'url': 'test-url-1',
                'mirror-of': 'test-mirror-of-1'
            },
            'ignore-me-2': {
                'id': 'test-id-2',
                'url': 'test-url-2',
                'mirror-of': 'test-mirror-of-2'
            }
        }

        add_maven_mirrors(
            parent_element=root_element,
            maven_mirrors=maven_mirrors
        )

        tree = ET.ElementTree(root_element)
        tree_write_out = BytesIO()
        tree.write(tree_write_out)

        self.assertEqual(
            tree_write_out.getvalue().decode(),
            r"<settings><mirrors>"
            r"<mirror>"
            r"<id>use-me-1</id>"
            r"<url>test-url-1</url>"
            r"<mirrorOf>test-mirror-of-1</mirrorOf>"
            r"</mirror>"
            r"<mirror>"
            r"<id>test-id-2</id>"
            r"<url>test-url-2</url>"
            r"<mirrorOf>test-mirror-of-2</mirrorOf>"
            r"</mirror>"
            r"</mirrors></settings>"
        )

    def test_add_maven_mirrors_list_missing_id(self):
        root_element = ET.Element('settings')

        maven_mirrors = [
            {
                'url': 'test-url-1',
                'mirror-of': 'test-mirror-of-1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven mirror must specify a 'id':"
            r" {'url': 'test-url-1', 'mirror-of': 'test-mirror-of-1'}"
        ):
            add_maven_mirrors(
                parent_element=root_element,
                maven_mirrors=maven_mirrors
            )

    def test_add_maven_mirrors_list_missing_url(self):
        root_element = ET.Element('settings')

        maven_mirrors = [
            {
                'id': 'test-id-1',
                'mirror-of': 'test-mirror-of-1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven mirror must specify a 'url':"
            r" {'id': 'test-id-1', 'mirror-of': 'test-mirror-of-1'}"
        ):
            add_maven_mirrors(
                parent_element=root_element,
                maven_mirrors=maven_mirrors
            )

    def test_add_maven_mirrors_list_missing_mirror_of(self):
        root_element = ET.Element('settings')

        maven_mirrors = [
            {
                'id': 'test-id-1',
                'url': 'test-url-1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven mirror must specify"
            r" a 'mirror-of': {'id': 'test-id-1', 'url': 'test-url-1'}"
        ):
            add_maven_mirrors(
                parent_element=root_element,
                maven_mirrors=maven_mirrors
            )

    def test_add_maven_mirrors_dict_missing_url(self):
        root_element = ET.Element('settings')

        maven_mirrors = {
            'ignore-me-1': {
                'id': 'test-id-1',
                'mirror-of': 'test-mirror-of-1'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven mirror \(ignore-me-1\) must specify a 'url':"
            r" {'id': 'test-id-1', 'mirror-of': 'test-mirror-of-1'}"
        ):
            add_maven_mirrors(
                parent_element=root_element,
                maven_mirrors=maven_mirrors
            )

    def test_add_maven_mirrors_dict_missing_mirror_of(self):
        root_element = ET.Element('settings')

        maven_mirrors = {
            'ignore-me-1': {
                'id': 'test-id-1',
                'url': 'test-url-1'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for maven mirror \(ignore-me-1\) must specify a 'mirror-of':"
            r" {'id': 'test-id-1', 'url': 'test-url-1'}"
        ):
            add_maven_mirrors(
                parent_element=root_element,
                maven_mirrors=maven_mirrors
            )

@patch('sh.mvn', create=True)
class TestMavenUtils_write_effective_pom(BaseTestCase):
    def test_success(self, mvn_mock):
        pom_file_path = 'input/pom.xml'
        effective_pom_path = '/tmp/output/effective-pom.xml'

        actual_effective_pom_path = write_effective_pom(
            pom_file_path=pom_file_path,
            output_path=effective_pom_path
        )
        self.assertEqual(actual_effective_pom_path, effective_pom_path)
        mvn_mock.assert_any_call(
            'help:effective-pom',
            f'-f={pom_file_path}',
            f'-Doutput={effective_pom_path}'
        )

    def test_with_one_profile_string(self, mvn_mock):
        pom_file_path = 'input/pom.xml'
        effective_pom_path = '/tmp/output/effective-pom.xml'

        actual_effective_pom_path = write_effective_pom(
            pom_file_path=pom_file_path,
            output_path=effective_pom_path,
            profiles='mock-profile1'
        )
        self.assertEqual(actual_effective_pom_path, effective_pom_path)
        mvn_mock.assert_any_call(
            'help:effective-pom',
            f'-f={pom_file_path}',
            f'-Doutput={effective_pom_path}',
            '-P', 'mock-profile1'
        )

    def test_with_one_profile_list(self, mvn_mock):
        pom_file_path = 'input/pom.xml'
        effective_pom_path = '/tmp/output/effective-pom.xml'

        actual_effective_pom_path = write_effective_pom(
            pom_file_path=pom_file_path,
            output_path=effective_pom_path,
            profiles=['mock-profile1']
        )
        self.assertEqual(actual_effective_pom_path, effective_pom_path)
        mvn_mock.assert_any_call(
            'help:effective-pom',
            f'-f={pom_file_path}',
            f'-Doutput={effective_pom_path}',
            '-P', 'mock-profile1'
        )

    def test_with_mutliple_profiles(self, mvn_mock):
        pom_file_path = 'input/pom.xml'
        effective_pom_path = '/tmp/output/effective-pom.xml'

        actual_effective_pom_path = write_effective_pom(
            pom_file_path=pom_file_path,
            output_path=effective_pom_path,
            profiles=['mock-profile1', 'mock-profile2']
        )
        self.assertEqual(actual_effective_pom_path, effective_pom_path)
        mvn_mock.assert_any_call(
            'help:effective-pom',
            f'-f={pom_file_path}',
            f'-Doutput={effective_pom_path}',
            '-P', 'mock-profile1,mock-profile2'
        )

    def test_fail(self, mvn_mock):
        pom_file_path = 'input/pom.xml'
        effective_pom_path = '/tmp/output/effective-pom.xml'

        mvn_mock.side_effect = sh.ErrorReturnCode('mvn', b'mock stdout', b'mock error')

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Error generating effective pom for '{pom_file_path}' to '{effective_pom_path}'"
                r".*RAN: mvn"
                r".*STDOUT:"
                r".*mock stdout"
                r".*STDERR:"
                r".*mock error",
                re.DOTALL
            )
        ):
            write_effective_pom(
                pom_file_path=pom_file_path,
                output_path=effective_pom_path
            )
            mvn_mock.assert_any_call(
                'help:effective-pom',
                f'-f={pom_file_path}',
                f'-Doutput={effective_pom_path}'
            )

    def test_fail_not_absolute_path(self, mvn_mock):
        pom_file_path = 'input/pom.xml'
        effective_pom_path = 'output/effective-pom.xml'

        mvn_mock.side_effect = sh.ErrorReturnCode('mvn', b'mock stdout', b'mock error')

        with self.assertRaisesRegex(
            StepRunnerException,
            re.compile(
                rf"Given output path \({effective_pom_path}\) is not absolute which will mean "
                rf"your output file will actually end up being relative to the pom file "
                rf"\({pom_file_path}\) rather than your expected root. Rather then handling this, "
                rf"just give this function an absolute path. "
                rf"If you are a user seeing this, a programmer messed up somewhere, "
                rf"report an issue.",
                re.DOTALL
            )
        ):
            write_effective_pom(
                pom_file_path=pom_file_path,
                output_path=effective_pom_path
            )
            mvn_mock.assert_any_call(
                'help:effective-pom',
                f'-f={pom_file_path}',
                f'-Doutput={effective_pom_path}'
            )

@patch('ploigos_step_runner.utils.maven.write_effective_pom')
class TestMavenUtils_get_effective_pom(BaseTestCase):
    def test_does_call_once(self, mock_write_effective_pom):
        with TempDirectory() as temp_dir:
            # set up mock
            def mock_write_effective_pom_side_effect(pom_file_path, output_path, profiles):
                Path(output_path).touch()
            mock_write_effective_pom.side_effect = mock_write_effective_pom_side_effect

            # run test
            effective_pom = get_effective_pom(
                work_dir_path=temp_dir.path,
                pom_file='mock-pom.xml',
                profiles=None
            )

            # validate
            expected_effective_pom = os.path.join(temp_dir.path, 'effective-pom.xml')
            self.assertEqual(effective_pom, expected_effective_pom)
            mock_write_effective_pom.assert_called_once_with(
                pom_file_path='mock-pom.xml',
                output_path=expected_effective_pom,
                profiles=None
            )

    def test_does_call_twice(self, mock_write_effective_pom):
        with TempDirectory() as temp_dir:
            # set up mock
            def mock_write_effective_pom_side_effect(pom_file_path, output_path, profiles):
                Path(output_path).touch()
            mock_write_effective_pom.side_effect = mock_write_effective_pom_side_effect

            # run test (first call)
            effective_pom = get_effective_pom(
                work_dir_path=temp_dir.path,
                pom_file='mock-pom.xml',
                profiles=None
            )

            # validate
            expected_effective_pom = os.path.join(temp_dir.path, 'effective-pom.xml')
            self.assertEqual(effective_pom, expected_effective_pom)
            mock_write_effective_pom.assert_called_once_with(
                pom_file_path='mock-pom.xml',
                output_path=expected_effective_pom,
                profiles=None
            )

            # run test (second call)
            mock_write_effective_pom.reset_mock()
            effective_pom = get_effective_pom(
                work_dir_path=temp_dir.path,
                pom_file='mock-pom.xml',
                profiles=None
            )

            # validate
            expected_effective_pom = os.path.join(temp_dir.path, 'effective-pom.xml')
            self.assertEqual(effective_pom, expected_effective_pom)
            mock_write_effective_pom.assert_not_called()

class TestMavenUtils_get_maven_plugin_xml_element_path(BaseTestCase):
    def test_given_plugin_name(self):
        actual_xml_element_path = get_maven_plugin_xml_element_path('maven-surefire-plugin')
        self.assertEqual(
            actual_xml_element_path,
            './mvn:build/mvn:plugins/mvn:plugin[mvn:artifactId="maven-surefire-plugin"]'
        )

class TestMavenUtils_run_maven(BaseTestCase):
    @patch('sh.mvn', create=True)
    @patch('ploigos_step_runner.utils.maven.create_sh_redirect_to_multiple_streams_fn_callback')
    @patch("builtins.open", new_callable=mock_open)
    def test_success_defaults(self, mock_open, redirect_mock, mvn_mock):
        with TempDirectory() as temp_dir:
            mvn_output_file_path = os.path.join(temp_dir.path, 'maven_output.txt')
            settings_file = '/fake/settings.xml'
            pom_file = '/fake/pom.xml'
            phases_and_goals = 'fake'

            run_maven(
                mvn_output_file_path=mvn_output_file_path,
                settings_file=settings_file,
                pom_file=pom_file,
                phases_and_goals=phases_and_goals,
            )

            mock_open.assert_called_with(mvn_output_file_path, 'w', encoding='utf-8')
            redirect_mock.assert_has_calls([
                call([
                    sys.stdout,
                    mock_open.return_value
                ]),
                call([
                    sys.stderr,
                    mock_open.return_value
                ])
            ])

            mvn_mock.assert_called_once_with(
                'fake',
                '-f', '/fake/pom.xml',
                '-s', '/fake/settings.xml',
                '--no-transfer-progress',
                _out=Any(StringIO),
                _err=Any(StringIO)
            )

    @patch('sh.mvn', create=True)
    @patch('ploigos_step_runner.utils.maven.create_sh_redirect_to_multiple_streams_fn_callback')
    @patch("builtins.open", new_callable=mock_open)
    def test_success_with_single_profile(self, mock_open, redirect_mock, mvn_mock):
        with TempDirectory() as temp_dir:
            mvn_output_file_path = os.path.join(temp_dir.path, 'maven_output.txt')
            settings_file = '/fake/settings.xml'
            pom_file = '/fake/pom.xml'
            phases_and_goals = 'fake'

            run_maven(
                mvn_output_file_path=mvn_output_file_path,
                settings_file=settings_file,
                pom_file=pom_file,
                phases_and_goals=phases_and_goals,
                profiles=['fake-profile']
            )

            mock_open.assert_called_with(mvn_output_file_path, 'w', encoding='utf-8')
            redirect_mock.assert_has_calls([
                call([
                    sys.stdout,
                    mock_open.return_value
                ]),
                call([
                    sys.stderr,
                    mock_open.return_value
                ])
            ])

            mvn_mock.assert_called_once_with(
                'fake',
                '-f', '/fake/pom.xml',
                '-s', '/fake/settings.xml',
                '-P', 'fake-profile',
                '--no-transfer-progress',
                _out=Any(StringIO),
                _err=Any(StringIO)
            )

    @patch('sh.mvn', create=True)
    @patch('ploigos_step_runner.utils.maven.create_sh_redirect_to_multiple_streams_fn_callback')
    @patch("builtins.open", new_callable=mock_open)
    def test_success_with_multiple_profile(self, mock_open, redirect_mock, mvn_mock):
        with TempDirectory() as temp_dir:
            mvn_output_file_path = os.path.join(temp_dir.path, 'maven_output.txt')
            settings_file = '/fake/settings.xml'
            pom_file = '/fake/pom.xml'
            phases_and_goals = 'fake'

            run_maven(
                mvn_output_file_path=mvn_output_file_path,
                settings_file=settings_file,
                pom_file=pom_file,
                phases_and_goals=phases_and_goals,
                profiles=['fake-profile1', 'fake-profile2']
            )

            mock_open.assert_called_with(mvn_output_file_path, 'w', encoding='utf-8')
            redirect_mock.assert_has_calls([
                call([
                    sys.stdout,
                    mock_open.return_value
                ]),
                call([
                    sys.stderr,
                    mock_open.return_value
                ])
            ])

            mvn_mock.assert_called_once_with(
                'fake',
                '-f', '/fake/pom.xml',
                '-s', '/fake/settings.xml',
                '-P', 'fake-profile1,fake-profile2',
                '--no-transfer-progress',
                _out=Any(StringIO),
                _err=Any(StringIO)
            )

    @patch('sh.mvn', create=True)
    @patch('ploigos_step_runner.utils.maven.create_sh_redirect_to_multiple_streams_fn_callback')
    @patch("builtins.open", new_callable=mock_open)
    def test_success_with_no_tls(self, mock_open, redirect_mock, mvn_mock):
        with TempDirectory() as temp_dir:
            mvn_output_file_path = os.path.join(temp_dir.path, 'maven_output.txt')
            settings_file = '/fake/settings.xml'
            pom_file = '/fake/pom.xml'
            phases_and_goals = 'fake'

            run_maven(
                mvn_output_file_path=mvn_output_file_path,
                settings_file=settings_file,
                pom_file=pom_file,
                phases_and_goals=phases_and_goals,
                tls_verify=False
            )

            mock_open.assert_called_with(mvn_output_file_path, 'w', encoding='utf-8')
            redirect_mock.assert_has_calls([
                call([
                    sys.stdout,
                    mock_open.return_value
                ]),
                call([
                    sys.stderr,
                    mock_open.return_value
                ])
            ])

            mvn_mock.assert_called_once_with(
                'fake',
                '-f', '/fake/pom.xml',
                '-s', '/fake/settings.xml',
                '--no-transfer-progress',
                '-Dmaven.wagon.http.ssl.insecure=true',
                '-Dmaven.wagon.http.ssl.allowall=true',
                '-Dmaven.wagon.http.ssl.ignore.validity.dates=true',
                _out=Any(StringIO),
                _err=Any(StringIO)
            )

    @patch('sh.mvn', create=True)
    @patch('ploigos_step_runner.utils.maven.create_sh_redirect_to_multiple_streams_fn_callback')
    @patch("builtins.open", new_callable=mock_open)
    def test_success_transfer_progress(self, mock_open, redirect_mock, mvn_mock):
        with TempDirectory() as temp_dir:
            mvn_output_file_path = os.path.join(temp_dir.path, 'maven_output.txt')
            settings_file = '/fake/settings.xml'
            pom_file = '/fake/pom.xml'
            phases_and_goals = 'fake'

            run_maven(
                mvn_output_file_path=mvn_output_file_path,
                settings_file=settings_file,
                pom_file=pom_file,
                phases_and_goals=phases_and_goals,
                no_transfer_progress=False
            )

            mock_open.assert_called_with(mvn_output_file_path, 'w', encoding='utf-8')
            redirect_mock.assert_has_calls([
                call([
                    sys.stdout,
                    mock_open.return_value
                ]),
                call([
                    sys.stderr,
                    mock_open.return_value
                ])
            ])

            mvn_mock.assert_called_once_with(
                'fake',
                '-f', '/fake/pom.xml',
                '-s', '/fake/settings.xml',
                None,
                _out=Any(StringIO),
                _err=Any(StringIO)
            )

    @patch('sh.mvn', create=True)
    @patch('ploigos_step_runner.utils.maven.create_sh_redirect_to_multiple_streams_fn_callback')
    @patch("builtins.open", new_callable=mock_open)
    def test_success_failure(self, mock_open, redirect_mock, mvn_mock):
        with TempDirectory() as temp_dir:
            mvn_output_file_path = os.path.join(temp_dir.path, 'maven_output.txt')
            settings_file = '/fake/settings.xml'
            pom_file = '/fake/pom.xml'
            phases_and_goals = 'fake'

            mvn_mock.side_effect = sh.ErrorReturnCode('mvn', b'mock stdout', b'mock error')

            with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error running maven."
                    r".*RAN: mvn"
                    r".*STDOUT:"
                    r".*mock stdout"
                    r".*STDERR:"
                    r".*mock error",
                    re.DOTALL
                )
            ):
                run_maven(
                    mvn_output_file_path=mvn_output_file_path,
                    settings_file=settings_file,
                    pom_file=pom_file,
                    phases_and_goals=phases_and_goals,
                )

                mock_open.assert_called_with(mvn_output_file_path, 'w', encoding='utf-8')
                redirect_mock.assert_has_calls([
                    call([
                        sys.stdout,
                        mock_open.return_value
                    ]),
                    call([
                        sys.stderr,
                        mock_open.return_value
                    ])
                ])

                mvn_mock.assert_called_once_with(
                    'fake',
                    '-f', '/fake/pom.xml',
                    '-s', '/fake/settings.xml',
                    '--no-transfer-progress',
                    _out=Any(StringIO),
                    _err=Any(StringIO)
                )

@patch('ploigos_step_runner.utils.maven.get_xml_element_text_by_path')
@patch('ploigos_step_runner.utils.maven.get_xml_element_by_path')
@patch('ploigos_step_runner.utils.maven.get_maven_plugin_xml_element_path')
@patch('ploigos_step_runner.utils.maven.get_effective_pom')
class TestMavenUtils_get_plugin_configuration_values(BaseTestCase):
    def test_no_profiles_no_phases(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        # set up mocks
        mock_get_xml_element_text_by_path.return_value = ['mock-config-value-1']

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, ['mock-config-value-1'])
        mock_get_xml_element_text_by_path.assert_called_once()
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )

    def test_with_profile_no_phases(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        # set up mocks
        mock_get_xml_element_text_by_path.return_value = ['mock-config-value-1']

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=['mock-profile'],
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, ['mock-config-value-1'])
        mock_get_xml_element_text_by_path.assert_called_once()
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=['mock-profile']
        )

    def test_with_no_profiles_one_phase_with_one_matching_config(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        mock_phases_with_values = ['mock-test-phase']

        # set up mocks
        def mock_get_xml_element_text_by_path_side_effect(
            xml_file_path,
            xpath,
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        ):
            for mock_phase in mock_phases_with_values:
                if f'[mvn:phase="{mock_phase}"]' in xpath:
                    return [f'{mock_phase}-mock-config-value']

        mock_get_xml_element_text_by_path.side_effect = mock_get_xml_element_text_by_path_side_effect

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=['mock-test-phase']
        )

        # validate
        self.assertEqual(actual_values, ['mock-test-phase-mock-config-value'])
        self.assertEqual(mock_get_xml_element_text_by_path.call_count, 2)
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )

    def test_with_no_profiles_one_phase_with_multiple_matching_config(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        mock_phases_with_values = ['mock-test-phase']

        # set up mocks
        def mock_get_xml_element_text_by_path_side_effect(
            xml_file_path,
            xpath,
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        ):
            for mock_phase in mock_phases_with_values:
                if f'[mvn:phase="{mock_phase}"]' in xpath:
                    return [f'{mock_phase}-mock-config-value-1', f'{mock_phase}-mock-config-value-2']

        mock_get_xml_element_text_by_path.side_effect = mock_get_xml_element_text_by_path_side_effect

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=['mock-test-phase']
        )

        # validate
        self.assertEqual(
            actual_values,
            ['mock-test-phase-mock-config-value-1', 'mock-test-phase-mock-config-value-2']
        )
        self.assertEqual(mock_get_xml_element_text_by_path.call_count, 2)
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )

    def test_with_no_profiles_multiple_phases_each_with_one_matching_config(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        mock_phases_with_values = ['mock-test-phase-1', 'mock-test-phase-2']

        # set up mocks
        def mock_get_xml_element_text_by_path_side_effect(
            xml_file_path,
            xpath,
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        ):
            for mock_phase in mock_phases_with_values:
                if f'[mvn:phase="{mock_phase}"]' in xpath:
                    return [f'{mock_phase}-mock-config-value']

        mock_get_xml_element_text_by_path.side_effect = mock_get_xml_element_text_by_path_side_effect

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=['mock-test-phase-1', 'mock-test-phase-2']
        )

        # validate
        self.assertEqual(
            actual_values,
            ['mock-test-phase-1-mock-config-value', 'mock-test-phase-2-mock-config-value']
        )
        self.assertEqual(mock_get_xml_element_text_by_path.call_count, 4)
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )

    def test_with_no_profiles_multiple_phases_each_with_multiple_matching_config(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        mock_phases_with_values = ['mock-test-phase-1', 'mock-test-phase-2']

        # set up mocks
        def mock_get_xml_element_text_by_path_side_effect(
            xml_file_path,
            xpath,
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        ):
            for mock_phase in mock_phases_with_values:
                if f'[mvn:phase="{mock_phase}"]' in xpath:
                    return [f'{mock_phase}-mock-config-value-1', f'{mock_phase}-mock-config-value-2']

        mock_get_xml_element_text_by_path.side_effect = mock_get_xml_element_text_by_path_side_effect

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=['mock-test-phase-1', 'mock-test-phase-2']
        )

        # validate
        self.assertEqual(
            actual_values,
            [
                'mock-test-phase-1-mock-config-value-1',
                'mock-test-phase-1-mock-config-value-2',
                'mock-test-phase-2-mock-config-value-1',
                'mock-test-phase-2-mock-config-value-2'
            ]
        )
        self.assertEqual(mock_get_xml_element_text_by_path.call_count, 4)
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )

    def test_with_no_profiles_multiple_phases_no_matching_config(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        # set up mocks
        mock_phases_without_values = ['mock-test-phase-1', 'mock-test-phase-2']
        def mock_get_xml_element_text_by_path_side_effect(
            xml_file_path,
            xpath,
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        ):
            for mock_phase in mock_phases_without_values:
                if f'[mvn:phase="{mock_phase}"]' in xpath:
                    return []

            return ['mock-default-value']

        mock_get_xml_element_text_by_path.side_effect = mock_get_xml_element_text_by_path_side_effect

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=['mock-test-phase-1', 'mock-test-phase-2']
        )

        # validate
        self.assertEqual(
            actual_values,
            ['mock-default-value']
        )
        self.assertEqual(mock_get_xml_element_text_by_path.call_count, 4)
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )

    def test_plugin_not_found(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        # set up mocks
        mock_get_xml_element_by_path.return_value = None

        # run test
        with self.assertRaisesRegex(
            RuntimeError,
            r"Expected maven plugin \(mock-maven-plugin\) not found in "
            r" effective pom for given pom \(mock-pom.xml\)."
        ):
            get_plugin_configuration_values(
                plugin_name='mock-maven-plugin',
                configuration_key='mockAwesomeConfig',
                work_dir_path='/mock/work_dir',
                pom_file='mock-pom.xml',
                profiles=None,
                phases_and_goals=['mock-test-phase-1', 'mock-test-phase-2']
            )

        # validate
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )
        mock_get_xml_element_text_by_path.assert_not_called()

    def test_with_no_profiles_one_goal_with_one_matching_config(
        self,
        mock_get_effective_pom,
        mock_get_maven_plugin_xml_element_path,
        mock_get_xml_element_by_path,
        mock_get_xml_element_text_by_path
    ):
        mock_goals_with_values = ['mock-test-goal']

        # set up mocks
        def mock_get_xml_element_text_by_path_side_effect(
            xml_file_path,
            xpath,
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        ):
            for mock_goal in mock_goals_with_values:
                if f'[mvn:goal="{mock_goal}"]' in xpath:
                    return [f'{mock_goal}-mock-config-value']

        mock_get_xml_element_text_by_path.side_effect = mock_get_xml_element_text_by_path_side_effect

        # run test
        actual_values = get_plugin_configuration_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=['mock-test-goal']
        )

        # validate
        self.assertEqual(actual_values, ['mock-test-goal-mock-config-value'])
        self.assertEqual(mock_get_xml_element_text_by_path.call_count, 2)
        mock_get_effective_pom.assert_called_once_with(
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None
        )
# Integration Test with XML, primarly for testing xpaths
# NOTE: isn't full integration test because mocking the effective pom so don't have to have
#       maven installed.
@patch('ploigos_step_runner.utils.maven.get_effective_pom')
class TestMavenUtils_get_plugin_configuration_values_IT_XML(BaseTestCase):

    def __get_test_file_path(self, file):
        return os.path.join(
            os.path.dirname(__file__),
            'files',
            'get_plugin_configuration_values_IT',
            file
        )

    def test_no_profiles_no_phases_no_config_surefire(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-no-plugin-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=None
            )

            # validate
            self.assertEqual(actual_values, [])

    def test_no_profiles_no_phases_no_config_failsafe(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-no-plugin-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-failsafe-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=None
            )

            # validate
            self.assertEqual(actual_values, [])

    def test_no_profiles_no_phases_with_config_surefire(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-plugin-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=None
            )

            # validate
            self.assertEqual(
                actual_values,
                ['${project.build.directory}/surefire-reports-unit-test']
            )

    def test_no_profiles_no_phases_with_config_failsafe(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-plugin-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-failsafe-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=None
            )

            # validate
            self.assertEqual(
                actual_values,
                ['${project.build.directory}/failsafe-reports-integration-test']
            )

    def test_no_profiles_surefire_both_UT_and_IT_via_phase_get_UT_config(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-plugin-phase-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=None
            )

            # validate
            self.assertEqual(
                actual_values,
                ['${project.build.directory}/surefire-reports-unit-test']
            )

    def test_no_profiles_surefire_both_UT_and_IT_via_phase_get_IT_config(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-plugin-phase-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=['integration-test']
            )

            # validate
            self.assertEqual(
                actual_values,
                ['${project.build.directory}/surefire-reports-uat']
            )

    def test_failsafe_with_default_and_goal_execution_config_specify_no_goal(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-plugin-goal-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-failsafe-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=[]
            )

            # validate
            self.assertEqual(
                actual_values,
                ['${project.build.directory}/failsafe-reports-default']
            )

    def test_failsafe_with_default_and_goal_execution_config_specify_goal(self, get_effective_pom_mock):
        # setup
        pom_file = self.__get_test_file_path('pom-plugin-goal-config.xml')
        get_effective_pom_mock.return_value = pom_file

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-failsafe-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles=None,
                phases_and_goals=['integration-test']
            )

            # validate
            self.assertEqual(
                actual_values,
                ['${project.build.directory}/failsafe-reports-execution']
            )

# Integration Test with XML and Maven
class TestMavenUtils_get_plugin_configuration_values_IT_Maven(BaseTestCase):

    def setUp(self):
        super().setUp()

        maven_path = sh.which('mvn')
        if not maven_path:
            self.skipTest('maven is not installed')

    def __get_test_file_path(self, file):
        return os.path.join(
            os.path.dirname(__file__),
            'files',
            'get_plugin_configuration_values_IT',
            file
        )

    def test_with_profile_no_phases_with_config_surefire(self):
        # setup
        pom_file = self.__get_test_file_path('pom-no-plugin-config.xml')

        with TempDirectory() as temp_dir:
            # run test
            actual_values = get_plugin_configuration_values(
                plugin_name='maven-surefire-plugin',
                configuration_key='reportsDirectory',
                work_dir_path=temp_dir.path,
                pom_file=pom_file,
                profiles='integration-test',
                phases_and_goals=None
            )

            # validate
            self.assertEqual(len(actual_values), 1)
            self.assertEqual(
                actual_values[0],
                os.path.join(
                    os.path.dirname(os.path.abspath(pom_file)),
                    'target',
                    'surefire-reports-integration-test'
                )
            )

@patch('ploigos_step_runner.utils.maven.get_plugin_configuration_values')
class TestMavenUtils_get_plugin_configuration_absolute_path_values(BaseTestCase):
    def test_one_relative_result_with_relative_pom(self, mock_get_plugin_configuration_values):
        # set up mock
        mock_get_plugin_configuration_values.return_value = ['mock/relative/path/foo1']

        # run test
        actual_values = get_plugin_configuration_absolute_path_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, [os.path.join(os.getcwd(), 'mock/relative/path/foo1')])

    def test_one_relative_result_with_absolute_pom(self, mock_get_plugin_configuration_values):
        # set up mock
        mock_get_plugin_configuration_values.return_value = ['mock/relative/path/foo1']

        # run test
        actual_values = get_plugin_configuration_absolute_path_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='/mock/absolute/path/mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, ['/mock/absolute/path/mock/relative/path/foo1'])

    def test_one_abslute_result_with_relative_pom(self, mock_get_plugin_configuration_values):
        # set up mock
        mock_get_plugin_configuration_values.return_value = ['/mock/absolute/path/foo1']

        # run test
        actual_values = get_plugin_configuration_absolute_path_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, ['/mock/absolute/path/foo1'])

    def test_one_abslute_result_with_absolute_pom(self, mock_get_plugin_configuration_values):
        # set up mock
        mock_get_plugin_configuration_values.return_value = ['/mock/absolute/path/foo1']

        # run test
        actual_values = get_plugin_configuration_absolute_path_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='/mock/awesome/different/aboslute/path/mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, ['/mock/absolute/path/foo1'])

    def test_no_config_values_found_none_returned(self, mock_get_plugin_configuration_values):
        # set up mock
        mock_get_plugin_configuration_values.return_value = None

        # run test
        actual_values = get_plugin_configuration_absolute_path_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, [])

    def test_no_config_values_found_empty_list_returned(self, mock_get_plugin_configuration_values):
        # set up mock
        mock_get_plugin_configuration_values.return_value = []

        # run test
        actual_values = get_plugin_configuration_absolute_path_values(
            plugin_name='mock-maven-plugin',
            configuration_key='mockAwesomeConfig',
            work_dir_path='/mock/work_dir',
            pom_file='mock-pom.xml',
            profiles=None,
            phases_and_goals=None
        )

        # validate
        self.assertEqual(actual_values, [])
